/*
 * large_sliced_coo.h
 *
 *  Created on: Dec 13, 2010
 *      Author: hans
 */

#ifndef LARGE_SLICED_COO_H_
#define LARGE_SLICED_COO_H_

#include "../util/util.h"
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <sstream>
#include "large_sliced_coo_kernel.cuh"
#include "../util/textures.h"

namespace cuda_bwc{

template<typename T>
class LargeSlicedCoo : public BinaryMatrix<T>{
public:
	LargeSlicedCoo() : BinaryMatrix<T>(0,0){};
	LargeSlicedCoo(uint32_t num_rows, uint32_t num_cols) :
		BinaryMatrix<T>(num_rows, num_cols){};

	void readMatrix(MatrixInput &in);
	void writeMatrix(std::ostream &out);
	void readCache(std::istream &in);
	void buildCache(std::ostream &out);
	std::string getCacheName();
	uint64_t memoryUsage();
	void multiply(const T *v, T *r);
	uint32_t granularity();
	uint32_t nonZeroes();
protected:
	virtual uint32_t numRowsPerSlice(){return NUM_ROWS_PER_SLICE;};
	uint32_t non_zeroes;
	thrust::host_vector<uint8_t> row;
    thrust::host_vector<uint32_t> col;
	thrust::host_vector<uint32_t> offsets;
    thrust::host_vector<T> value;

	thrust::device_vector<uint8_t> dv_row;
    thrust::device_vector<uint32_t> dv_col;
	thrust::device_vector<uint32_t> dv_offsets;
    thrust::device_vector<T> dv_value;
	static const uint32_t GROUP_SIZE = 1<<30;
    void transferToDevice();

private:
	static const uint32_t THREADS_PER_BLOCK = 1024;
	static const uint32_t NUM_ROWS_PER_SLICE = SHARED_MEMORY_SIZE/(WARP_SIZE*sizeof(T))>256? 256 :
												SHARED_MEMORY_SIZE/(WARP_SIZE*sizeof(T));

	void pushCoo(uint16_t n);


};

template <typename T>
std::string LargeSlicedCoo<T>::getCacheName(){
	std::stringstream ss;
	ss << "-lscoo-" << NUM_ROWS_PER_SLICE << "pack-" << this->num_rows << "-" << this->num_cols << ".bin";
	return ss.str();
}

template<typename T>
void LargeSlicedCoo<T>::pushCoo(uint16_t n){
	//coo.push_back(n);
	//coo.push_back(n>>8);
}

template<typename T>
void LargeSlicedCoo<T>::readMatrix(MatrixInput &in){
	vector < pair<uint32_t, pair< uint16_t, double  > > > slice;
	uint32_t nz = 0;
	uint32_t numGroups = divideAndCeil(this->num_cols, GROUP_SIZE);
	uint32_t maxGroup = numGroups * GROUP_SIZE;
	for (uint32_t i = 0; i < this->num_rows;) {
		uint32_t limit = min(this->num_rows, i + numRowsPerSlice());
		slice.clear();
		for (; i < limit; i++) {
			uint32_t temp;
			in >> temp;
			nz += temp;
			for (uint32_t j = 0; j < temp; j++) {
				uint32_t c;
				in >> c;
                double v;
                v = in.getValue();
				slice.push_back(make_pair(c, make_pair( static_cast<uint8_t>(i % numRowsPerSlice()), v )) );
			}
		}
		sort(slice.begin(), slice.end());
		offsets.push_back(row.size());
		uint32_t curGroup = GROUP_SIZE;
		for (uint32_t j = 0; j < slice.size(); j++) {
			uint32_t c = slice[j].first;
			uint8_t r = slice[j].second.first;
            double v = slice[j].second.second;
			while(c >= curGroup){
				curGroup += GROUP_SIZE;
				offsets.push_back(row.size());
			}
		//	pushCoo(c-(curGroup-GROUP_SIZE));
		//	coo.push_back(r);
		    row.push_back(r);
            col.push_back(c - (curGroup-GROUP_SIZE) );
            value.push_back((T) v); 
		}
		while(curGroup<maxGroup){
			curGroup += GROUP_SIZE;
			offsets.push_back(row.size());
		}
	}
	offsets.push_back(row.size());
	non_zeroes = nz;
    this->setStat( (double)nz / (double)THREADS_PER_BLOCK / this->num_rows );
//	transferToDevice();
}

template<typename T>
void LargeSlicedCoo<T>::writeMatrix(std::ostream &out){
	/*vector<uint32_t> rows[numRowsPerSlice()];
	uint32_t last = ((this->num_rows-1) % numRowsPerSlice()) + 1;
	uint32_t numPacks = divideAndCeil(this->num_rows, numRowsPerSlice());
	uint32_t numGroups = divideAndCeil(this->num_cols, GROUP_SIZE);
	for (uint32_t i = 0; i < numPacks; i++) {
		uint32_t baseVal = 0;
		uint32_t limit = (i == (numPacks - 1)) ? last : numRowsPerSlice();
		for(uint32_t g=0; g<numGroups; g++){
			for (uint32_t j = offsets[i*numGroups+g]; j < offsets[i*numGroups+g+1]; j+=3) {
				uint32_t col = combine(coo[j], coo[j+1]) + baseVal;
				uint8_t row = coo[j+2];
				rows[row].push_back(col);
			}
			baseVal += GROUP_SIZE;
		}
		for (uint32_t j = 0; j < limit; j++) {
			out << rows[j].size();
			for (uint32_t k = 0; k < rows[j].size(); k++) {
				out << " " << rows[j][k];
			}
			rows[j].clear();
			out << endl;
		}
	}*/
}

template<typename T>
void LargeSlicedCoo<T>::readCache(std::istream &in){
	DataInputStream dis(in);
	dis >> (this->num_rows) >> (this->num_cols) >> (non_zeroes);
	dis.readVector(col);
    dis.readVector(row);
    dis.readVector(value);
	dis.readVector(offsets);

	transferToDevice();
}

template<typename T>
void LargeSlicedCoo<T>::buildCache(std::ostream &out){
	DataOutputStream dos(out);
	dos << this->num_rows << this->num_cols << non_zeroes;
	dos.writeVector(col);
    dos.writeVector(row);
    dos.writeVector(value);
	dos.writeVector(offsets);
}

template <typename T>
void LargeSlicedCoo<T>::transferToDevice(){
	dv_col.assign(col.begin(), col.end());
    dv_row.assign(row.begin(), row.end());
    dv_value.assign(value.begin(), value.end());
	dv_offsets.assign(offsets.begin(), offsets.end());
}

template<typename T>
void LargeSlicedCoo<T>::multiply(const T *v, T *r){
//	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
	large_sliced_coo_kernel_32<T, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE), THREADS_PER_BLOCK>>>(
					this->num_rows,
	                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
	                divideAndCeil(this->num_cols, GROUP_SIZE),
	                GROUP_SIZE,
	                thrust::raw_pointer_cast(&dv_col[0]),
                    thrust::raw_pointer_cast(&dv_row[0]),
                    thrust::raw_pointer_cast(&dv_value[0]),
	                thrust::raw_pointer_cast(&dv_offsets[0]),
	                v,
	                r);
}

/*template<>
void LargeSlicedCoo<uint32_t>::multiply(const uint32_t *v, uint32_t *r){
	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
	large_sliced_coo_kernel_32<uint32_t, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(
					this->num_rows,
	                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
	                divideAndCeil(this->num_cols, GROUP_SIZE),
	                GROUP_SIZE,
	                thrust::raw_pointer_cast(&dv_coo[0]),
                    thrust::raw_pointer_cast(&dv_value[0]),
	                thrust::raw_pointer_cast(&dv_offsets[0]),
	                v,
	                r);
}*/

template<typename T>
uint64_t LargeSlicedCoo<T>:: memoryUsage(){
	return col.size()*sizeof(col[0]) + row.size()*sizeof(row[0]) + offsets.size()*sizeof(offsets[0]) + value.size() * sizeof(value[0]);
}

template<typename T>
uint32_t LargeSlicedCoo<T>::granularity(){
	return numRowsPerSlice();
}

template<typename T>
uint32_t LargeSlicedCoo<T>::nonZeroes(){
	return non_zeroes;
}

}

#endif /* LARGE_SLICED_COO_H_ */
