/*
 * xl_sliced_coo.h
 *
 *  Created on: Dec 20, 2010
 *      Author: hans
 */

#ifndef XL_SLICED_COO_H_
#define XL_SLICED_COO_H_

#include "../util/util.h"
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <sstream>
#include "xl_sliced_coo_kernel.cuh"
#include "../util/textures.h"
#include <math.h>

namespace cuda_bwc{

template<typename T>
class XLSlicedCoo : public BinaryMatrix<T>{
public:
	XLSlicedCoo() : BinaryMatrix<T>(0,0){};
	XLSlicedCoo(uint32_t num_rows, uint32_t num_cols) :
		BinaryMatrix<T>(num_rows, num_cols){};

	void readMatrix(MatrixInput &in);
	void writeMatrix(std::ostream &out);
	void readCache(std::istream &in);
	void buildCache(std::ostream &out);
	std::string getCacheName();
	uint64_t memoryUsage();
	void multiply(const T *v, T *r);
	uint32_t granularity();
	uint32_t nonZeroes();
protected:
	virtual uint32_t numRowsPerSlice(){return NUM_ROWS_PER_SLICE;};
    virtual uint32_t GROUP_SIZE() { return __GROUP_SIZE; };
    virtual uint32_t getLaneSize() { return 1; };
	uint32_t non_zeroes;
	thrust::host_vector<uint16_t> row;
    thrust::host_vector<uint32_t> col;

	thrust::host_vector<uint32_t> offsets;
    thrust::host_vector<T> value;

	thrust::device_vector<uint32_t> dv_col;
    thrust::device_vector<uint16_t> dv_row; 
	thrust::device_vector<uint32_t> dv_offsets;
    thrust::device_vector<T> dv_value;

	void transferToDevice();

private:
	static const uint32_t THREADS_PER_BLOCK = 1024;
	static const uint32_t NUM_ROWS_PER_SLICE = SHARED_MEMORY_SIZE/sizeof(T);
	static const uint32_t __GROUP_SIZE = static_cast<uint32_t>(1<<31) ;

};

template <typename T>
std::string XLSlicedCoo<T>::getCacheName(){
	std::stringstream ss;
	ss << "-xlscoo-" << NUM_ROWS_PER_SLICE << "pack-" << this->num_rows << "-" << this->num_cols << ".bin";
	return ss.str();
}

template<typename T>
void XLSlicedCoo<T>::readMatrix(MatrixInput &in){
	vector < pair<uint32_t, pair< uint16_t, T > > > slice;
	uint32_t nz = 0;
	uint32_t numGroups = divideAndCeil(this->num_cols, GROUP_SIZE());
	uint32_t maxGroup = numGroups * GROUP_SIZE();

    double dist=0,ndist=0;
    
	for (uint32_t i = 0; i < this->num_rows;) {
		uint32_t limit = min(this->num_rows, i + numRowsPerSlice());
		slice.clear();
		for (; i < limit; i++) {
			uint32_t temp;
			in >> temp;
			nz += temp;
			for (uint32_t j = 0; j < temp; j++) {
				uint32_t c;
				in >> c;
                T v;
                v = (T) in.getValue();
				slice.push_back(make_pair(c, make_pair( static_cast<uint16_t>(i % numRowsPerSlice() ), v  )));
			}
		}
		sort(slice.begin(), slice.end());

		offsets.push_back(col.size());
		uint32_t curGroup = GROUP_SIZE();
		for (uint32_t j = 0; j < slice.size(); j++) {
			uint32_t c = slice[j].first;
			uint16_t r = slice[j].second.first;
            T v = slice[j].second.second;
			while(c >= curGroup){
				curGroup += GROUP_SIZE();
				offsets.push_back(col.size());
                cerr << "FAILED ! column index larger than 2^31 ?" << endl;
                exit(-1);
			}

            if (col.size() > 0) {
                dist += 1 / (1+std::abs((double) col.back() - c));
                ndist++;
            }
            
            col.push_back(c - (curGroup - GROUP_SIZE()) );
            row.push_back(r);
            value.push_back(v);

           
		}
		while(curGroup<maxGroup){
			curGroup += GROUP_SIZE();
			offsets.push_back(col.size());
		}
	}
/*    double t=0, nt=0, maxlocal=0;
    for (int i=0;i<col.size();) {
        int j = i;
        vector<bool> checked(1024,false);
        for (;i<(j+1024);i++) {
            maxlocal=0;
            if(!checked[i-j]) { 
                int localt = 1;        
                for (int k=i+1;k<(j+1024);k++) {
                    if (row[k] == row[i]) {
                        checked[k-j]=true;
                        localt++;
                    }
                }
                if (localt > maxlocal) maxlocal = localt;
            }
            t+= 1/( 1+maxlocal/getLaneSize() );
            nt++;
        } 
        t+=localt;
    }*/
	offsets.push_back(col.size());
	non_zeroes = nz;
//    this->setStat((double) nz*1e9 / (double) this->num_rows / (double) this->num_cols );
    this->setStat( ndist / dist );
//    this->setStat( (double)nz/getLaneSize()/(this->num_rows) );
	//transferToDevice();
}

template<typename T>
void XLSlicedCoo<T>::writeMatrix(std::ostream &out){
/*	vector<uint32_t> rows[numRowsPerSlice()];
	uint32_t last = ((this->num_rows-1) % numRowsPerSlice()) + 1;
	uint32_t numPacks = divideAndCeil(this->num_rows, numRowsPerSlice());
	uint32_t numGroups = divideAndCeil(this->num_cols, GROUP_SIZE);
	for (uint32_t i = 0; i < numPacks; i++) {
		uint32_t baseVal = 0;
		uint32_t limit = (i == (numPacks - 1)) ? last : numRowsPerSlice();
		for(uint32_t g=0; g<numGroups; g++){
			for (uint32_t j = offsets[i*numGroups+g]; j < offsets[i*numGroups+g+1]; j+=2) {
				uint32_t col = coo[j] + baseVal;
				uint16_t row = coo[j+1];
				rows[row].push_back(col);
			}
			baseVal += GROUP_SIZE;
		}
		for (uint32_t j = 0; j < limit; j++) {
			out << rows[j].size();
			for (uint32_t k = 0; k < rows[j].size(); k++) {
				out << " " << rows[j][k];
			}
			rows[j].clear();
			out << endl;
		}
	} */
}

template<typename T>
void XLSlicedCoo<T>::readCache(std::istream &in){
	DataInputStream dis(in);
	dis >> (this->num_rows) >> (this->num_cols) >> (non_zeroes);
	dis.readVector(col);
    dis.readVector(row);
    dis.readVector(value);
	dis.readVector(offsets);

	transferToDevice();
}

template<typename T>
void XLSlicedCoo<T>::buildCache(std::ostream &out){
	DataOutputStream dos(out);
	dos << this->num_rows << this->num_cols << non_zeroes;
	dos.writeVector(col);
    dos.writeVector(row);
    dos.writeVector(value);
	dos.writeVector(offsets);
}

template <typename T>
void XLSlicedCoo<T>::transferToDevice(){
	//dv_coo.resize(coo.size());
	//cudaMemcpy(thrust::raw_pointer_cast(&dv_coo[0]), &coo[0], sizeof(coo[0])*coo.size(), cudaMemcpyHostToDevice);
//	dv_coo.assign(coo.begin(), coo.end());
    dv_col.assign(col.begin(), col.end());
    dv_row.assign(row.begin(), row.end());
    dv_value.assign(value.begin(), value.end());
	dv_offsets.assign(offsets.begin(), offsets.end());

}

/*
template<>
void XLSlicedCoo<uint32_t>::multiply(const uint32_t *v, uint32_t *r){
	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
	xl_sliced_coo_kernel_32<uint32_t, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(
						this->num_rows,
		                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
		                divideAndCeil(this->num_cols, GROUP_SIZE),
		                GROUP_SIZE,
		                thrust::raw_pointer_cast(&dv_coo[0]),
                        thurst::raw_pointer_cast(&dv_value[0]),
		                thrust::raw_pointer_cast(&dv_offsets[0]),
		                v,
		                r);
}*/

template<typename T>
void XLSlicedCoo<T>::multiply(const T *v, T *r){
//	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
    bind_x(v);
	xl_sliced_coo_kernel_32<T, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE), THREADS_PER_BLOCK>>> (//dim3(THREADS_PER_BLOCK/4, 4)>>>(
						this->num_rows,
						divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
						divideAndCeil(this->num_cols, GROUP_SIZE()),
						GROUP_SIZE(),
						thrust::raw_pointer_cast(&dv_col[0]),
                        thrust::raw_pointer_cast(&dv_row[0]),
                        thrust::raw_pointer_cast(&dv_value[0]),
						thrust::raw_pointer_cast(&dv_offsets[0]),
						v,
						r);
    unbind_x(v);
}

//template<>
//void XLSlicedCoo<uint32_t>::multiply(const uint32_t *v, uint32_t *r){
//	static const uint32_t NUM_BLOCKS = getNumMultiprocessors();
//	xl_sliced_coo_kernel_32<uint32_t, THREADS_PER_BLOCK, NUM_ROWS_PER_SLICE> <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(
//					this->num_rows,
//	                divideAndCeil(this->num_rows, NUM_ROWS_PER_SLICE),
//	                divideAndCeil(this->num_cols, GROUP_SIZE),
//	                GROUP_SIZE,
//	                thrust::raw_pointer_cast(&dv_coo[0]),
//	                thrust::raw_pointer_cast(&dv_offsets[0]),
//	                v,
//	                r);
//}

template<typename T>
uint64_t XLSlicedCoo<T>:: memoryUsage(){
	return col.size()*sizeof(col[0]) + row.size()*sizeof(row[0]) + offsets.size()*sizeof(offsets[0]) + value.size()*sizeof(value[0]);
}

template<typename T>
uint32_t XLSlicedCoo<T>::granularity(){
	return numRowsPerSlice();
}

template<typename T>
uint32_t XLSlicedCoo<T>::nonZeroes(){
	return non_zeroes;
}

}

#endif /* XL_SLICED_COO_H_ */
