/*
 * xl_sliced_coo_kernel.h
 *
 *  Created on: Dec 20, 2010
 *      Author: hans
 */

#ifndef XL_SLICED_COO_KERNEL_H_
#define XL_SLICED_COO_KERNEL_H_

#include <stdint.h>
#include <stdio.h>
#include "../util/util.h"
#include "../util/textures.h"
#include "../util/xor.h"

namespace cuda_bwc{

template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE, const int LANE_SIZE>
__global__ void
sliced_coo_kernel_32(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t * cols,
                const uint16_t * rows, 
                const ValueType * V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
    const int thread_lane = threadIdx.x & (LANE_SIZE-1);
    const int row_lane = threadIdx.x/(LANE_SIZE);
 
    __shared__ ValueType sdata[NUM_ROWS_PER_SLICE][LANE_SIZE];
    
    const int packNo=blockIdx.x;
	const int limit = ( (packNo==numPacks-1)?((num_rows-1)%NUM_ROWS_PER_SLICE)+1:NUM_ROWS_PER_SLICE );

    const uint32_t begin = offsets[packNo];
    const uint32_t end = offsets[packNo+1];
    for(int i=row_lane; i<limit; i+=THREADS_PER_BLOCK/LANE_SIZE){
        sdata[i][thread_lane] = 0;
    }
    
    __syncthreads();

    for(int index=begin+threadIdx.x; index<end; index+=THREADS_PER_BLOCK){
        const uint32_t col = cols[index];
        const uint16_t row = rows[index];
        const ValueType value = V[index];

        const ValueType input = fetch_x(col, x) * value;
        atomic_add(&sdata[row][thread_lane], input);
    }

    __syncthreads();

    for(int j=LANE_SIZE/2; j>0; j>>=1){   
        for(uint32_t i=row_lane; i<limit; i+=THREADS_PER_BLOCK/LANE_SIZE)
            if(thread_lane < j)
                sdata[i][thread_lane] += sdata[i][thread_lane + j];
        __syncthreads();
    }

    const uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;

    for(int r = threadIdx.x; r < limit; r+=THREADS_PER_BLOCK){
        y[actualRow + r] = sdata[r][0];
    }
}


template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE>
__global__ void
xl_sliced_coo_kernel_32(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t numGroups,
                const uint32_t groupSize,
                const uint32_t * cols,
                const uint16_t * rows,
                const ValueType * V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
    __shared__ ValueType sdata[NUM_ROWS_PER_SLICE];
    
    const uint32_t packNo=blockIdx.x;
	const uint32_t limit = ( (packNo==numPacks-1)?((num_rows-1)%NUM_ROWS_PER_SLICE)+1:NUM_ROWS_PER_SLICE );

    for(uint32_t i=threadIdx.x; i<limit; i+=THREADS_PER_BLOCK){
        sdata[i] = 0;
    }
    syncthreads();

    uint32_t groupVal = 0;
    const uint32_t base_offset = packNo*numGroups;
    int32_t begin = offsets[base_offset];
    int32_t end = offsets[base_offset+1];
    for(int32_t index=begin+threadIdx.x; index<end; index+=THREADS_PER_BLOCK){
        const uint32_t col = cols[index] + groupVal;
        const uint32_t row = rows[index];
        const ValueType value = V[index];

        const ValueType input = fetch_x(col, x) * value;
        atomic_add(&sdata[row], input);
    }
    syncthreads();

    const uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;

    for(uint32_t r = threadIdx.x; r < limit; r+=THREADS_PER_BLOCK){
        y[actualRow+r] = sdata[r];
    }
}

template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE, const uint32_t OVER_SIZE>
__global__ void
ad_sliced_coo_kernel(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t numGroups,
                const uint32_t groupSize,
                const uint32_t * cols,
                const uint16_t * rows,
                const ValueType *V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
	const uint32_t NUM_ELEMS = NUM_ROWS_PER_SLICE / OVER_SIZE; 
    __shared__ ValueType sdata[NUM_ELEMS];
    
//    ValueType *pResult;

    const uint32_t packNo=blockIdx.x;
    const uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;
    const uint32_t limit = ( (packNo==numPacks-1)?((num_rows-1)%NUM_ROWS_PER_SLICE)+1:NUM_ROWS_PER_SLICE );

    for(uint32_t i=threadIdx.x; i<limit && i<NUM_ELEMS; i+=THREADS_PER_BLOCK){
        sdata[i] = 0;
    }
    syncthreads();

    uint32_t groupVal = 0;
    const uint32_t base_offset = packNo*numGroups;
    int32_t begin = offsets[base_offset];
    int32_t end;

    for(uint32_t i=1; i<=numGroups; i++){
        end = offsets[base_offset+i];
        for(int32_t index=begin+threadIdx.x; index<end; index+=THREADS_PER_BLOCK){
            uint32_t col = cols[index] + groupVal;
            uint32_t row = rows[index];

            ValueType input = fetch_x(col, x) * V[index];
            if (row < NUM_ELEMS)
            //    pResult = &sdata[row];
              atomic_add(&sdata[row], input);
            else
              atomic_add(&y[actualRow + row],input);
            //    pResult = &y[actualRow + row];
            //atomic_add(pResult , input);
        }
        groupVal += groupSize;
        begin = end;
    }

    syncthreads();

    for(uint32_t r = threadIdx.x; r < limit && r<NUM_ELEMS; r+=THREADS_PER_BLOCK){
        y[actualRow+r] = sdata[r];
    }

}

template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE>
__global__ void
xl_sliced_coo_kernel(
				const uint32_t num_rows,
                const uint32_t numPacks,
                const uint32_t numGroups,
                const uint32_t groupSize,
                const uint32_t * coo,
                const ValueType *V,
                const uint32_t * offsets,
                const ValueType * x,
                      ValueType * y)
{
	const uint32_t thread_id = blockDim.x * threadIdx.y + threadIdx.x;
    __shared__ ValueType sdata[NUM_ROWS_PER_SLICE];

    const uint32_t packNo=blockIdx.x;
    const uint32_t limit = ( (packNo==numPacks-1)?((num_rows-1)%NUM_ROWS_PER_SLICE)+1:NUM_ROWS_PER_SLICE );

        syncthreads();
        for(uint32_t i=thread_id; i<NUM_ROWS_PER_SLICE; i+=THREADS_PER_BLOCK){
            sdata[i] = 0;
        }
        syncthreads();

        uint32_t groupVal = groupSize*threadIdx.y;
        const uint32_t base_offset = packNo*numGroups;

        for(uint32_t i=threadIdx.y; i<numGroups; i+=blockDim.y){
        	int32_t begin = offsets[base_offset+i];
        	int32_t end = offsets[base_offset+i+1];
        	for(int32_t index=begin+threadIdx.x*2; index<end; index+=blockDim.x*2){
        		uint32_t col = coo[index] + groupVal;
				uint16_t row = coo[index+1];
				ValueType input = fetch_x(col, x); 
				uint32_t start = (row&~31)  + (row&31);
                atomic_add(&sdata[start], input);
        	}
        	groupVal += groupSize*blockDim.y;
        }

        syncthreads();

        const uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;

        for(uint32_t r = thread_id; r < limit; r+=THREADS_PER_BLOCK){
        	uint32_t start = (r&~31) + (r&31);
            y[actualRow+r] = sdata[start];
        }
}

//template <typename ValueType, const uint32_t THREADS_PER_BLOCK, const uint32_t NUM_ROWS_PER_SLICE>
//__global__ void
//xl_sliced_coo_kernel_64_128_256(
//				const uint32_t num_rows,
//                const uint32_t numPacks,
//                const uint32_t numGroups,
//                const uint32_t groupSize,
//                const uint16_t * coo,
//                const uint32_t * offsets,
//                const ValueType * x,
//                      ValueType * y)
//{
//    const uint32_t REPEAT = sizeof(ValueType)/sizeof(uint32_t);
//    __shared__ uint32_t sdata[REPEAT][NUM_ROWS_PER_SLICE];
//
//    uint32_t packNo;
//	uint32_t limit = NUM_ROWS_PER_SLICE;
//	
//    for(packNo=blockIdx.x; packNo<numPacks; packNo+=gridDim.x){
//    	if(packNo == numPacks-1){
//    		limit = ((num_rows-1)%NUM_ROWS_PER_SLICE)+1;
//    	}
//
//    	syncthreads();
//        for(uint32_t i=row_lane; i<limit; i+=THREADS_PER_BLOCK){
//        	scattered_write<uint32_t,REPEAT>(&sdata[0][i], NUM_ROWS_PER_SLICE, 0);
//        }
//        syncthreads();
//        
//        uint32_t groupVal = 0;
//        const uint32_t base_offset = packNo*numGroups;
//        int32_t begin = offsets[base_offset];
//        int32_t end;
//
//        for(uint32_t i=1; i<=numGroups; i++){
//        	end = offsets[base_offset+i];
//        	for(int32_t index=begin+threadIdx.x*3; index<end-2; index+=THREADS_PER_BLOCK*3){
//        		uint32_t col = combine(coo[index], coo[index+1]) + groupVal;
//				uint16_t row = coo[index+2];
//				ValueType input = fetch_x(col, x);
//				atomic_xor(&sdata[row][0][thread_lane], input, 32);
//        	}
//        	groupVal += groupSize;
//        	begin = end;
//        }
//
//        syncthreads();
//
//        for(uint32_t j=16; j>=1; j>>=1){
//            for(uint32_t i=row_lane; i<limit; i+=THREADS_PER_BLOCK/32){
//                if(thread_lane < j){
//                	XOR<REPEAT>(&sdata[i][0][(thread_lane+i)&(31)], &sdata[i][0][(thread_lane+i+j)&(31)], 32);
//                }
//            }
//            syncthreads();
//        }
//
//        uint32_t actualRow = packNo * NUM_ROWS_PER_SLICE;
//
//        for(uint32_t id=threadIdx.x; id<limit*REPEAT; id+=THREADS_PER_BLOCK){
//            reinterpret_cast<uint32_t*>(&y[actualRow+id/REPEAT])[id&(REPEAT-1)] = sdata[id/REPEAT][id&(REPEAT-1)][(id/REPEAT)&31];
//        }
//
//    }
//}

}

#endif /* XL_SLICED_COO_KERNEL_H_ */
